from ._hamming_codec import (  # noqa: F401
    encode,
    decode,
    ParityLocation,
    # n_parity_bits,
    # compute_parity_bit_positions,
    # compute_parity_bits,
    # n_parity_bits_required,
    # int2bin,
)
from ._hamming_codec import __version__  # noqa: F401
